* xpkBLZW - XPK interface to LZW compressor/decompressor
* Copyright (C) 1992 Bryan Ford
*
* This file is part of BLZW, Bryan's LZW compressor.
*
* BLZW is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* BLZW is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


        include "exec/types.i"
        include "exec/initializers.i"
        include "exec/libraries.i"
        include "exec/lists.i"
        include "exec/alerts.i"
        include "exec/resident.i"
        include "exec/semaphores.i"
        include "exec/funcdef.i"
        include "exec/exec_lib.i"
        include "utility/hooks.i"
        include "utility/tagitem.i"
        include "libraries/xpk.i"
        include "libraries/xpksub.i"
        include "bry/macros.i"

VERSION equ     3               ; Don't forget to change the idstring
REVISION equ    0


                rsreset
xsp_blah        rs.b    xsp_Sub
xsp_CompTab     rs.l    1
xsp_CompTabSize rs.l    1

                rsreset
clib_Node       rs.b    LIB_SIZE
clib_Flags      rs.b    1       ; Defined below
clib_pad        rs.b    1
clib_SysBase    rs.l    1       ; Pointer to ExecBase
clib_SegList    rs.l    1       ; Library SegList
clib_Size       rs.b    0

        BITDEF  CLIB,DELEXP,0           ; Delayed expunge

        xref    GenCompress,GenCompTableSize
        xref    GenDecompress,GenDecompTableSize

        code    text

dosentry
        moveq   #-1,d0
        rts

idstring
        dc.b    "xpkBLZW.library V3.0 - Copyright 1992 Bryan Ford",0
        even

initdescrip
        dc.w    RTC_MATCHWORD   ; UWORD RT_MATCHWORD
        dc.l    initdescrip     ; APTR  RT_MATCHTAG
        dc.l    endcode         ; APTR  RT_ENDSKIP
        dc.b    RTF_AUTOINIT    ; UBYTE RT_FLAGS
        dc.b    VERSION         ; UBYTE RT_VERSION
        dc.b    NT_LIBRARY      ; UBYTE RT_TYPE
        dc.b    0               ; BYTE  RT_PRI
        dc.l    libname         ; APTR  RT_NAME
        dc.l    idstring        ; APTR  RT_IDSTRING
        dc.l    inittab         ; APTR  RT_INIT

inittab
        dc.l    clib_Size       ; size of library base data space
        dc.l    functable       ; pointer to function initializers
        dc.l    datatable       ; pointer to data initializers
        dc.l    initroutine     ; routine to run

functable
        dc.l    Open
        dc.l    Close
        dc.l    Expunge
        dc.l    Null

        dc.l    packerinfo              ; PackerInfo
        dc.l    packchunk               ; PackChunk
        dc.l    freebuf                 ; PackFree
        dc.l    Null                    ; PackReset
        dc.l    unpackchunk             ; UnpackChunk
        dc.l    freebuf                 ; UnpackFree

        dc.l    -1

datatable:
        INITBYTE        LN_TYPE,NT_LIBRARY
        INITLONG        LN_NAME,libname
        INITBYTE        LIB_FLAGS,LIBF_SUMUSED!LIBF_CHANGED
        INITWORD        LIB_VERSION,VERSION
        INITWORD        LIB_REVISION,REVISION
        INITLONG        LIB_IDSTRING,idstring
        dc.l   0

initroutine:
        move.l  d0,a1
        move.l  a6,(clib_SysBase,a1)
        move.l  a0,(clib_SegList,a1)
        rts

Open       ; ( libptr:a6, version:d0 )
        addq.w  #1,(LIB_OPENCNT,a6)
        bclr    #CLIBB_DELEXP,(clib_Flags,a6)
        move.l  a6,d0
        rts

Close      ; ( libptr:a6 )
        subq.w  #1,(LIB_OPENCNT,a6)
        bnz.b   \out

        btst    #CLIBB_DELEXP,(clib_Flags,a6)
        bnz.b   Expunge
\out
        moveq   #0,d0
        rts

Expunge    ; ( libptr: a6 )
        movem.l d2/a5/a6,-(sp)
        move.l  a6,a5
        move.l  (clib_SysBase,a5),a6

        ;------ see if anyone has us open
        tst.w   (LIB_OPENCNT,a5)
        bz.b    \die

        ;------ it is still open.  set the delayed expunge flag
        bset    #CLIBB_DELEXP,(clib_Flags,a5)
        moveq   #0,d0
        bra.b   \out

\die
        ;------ go ahead and get rid of us.  Store our seglist in d2
        move.l  (clib_SegList,a5),d2

        ;------ unlink from library list
        move.l  a5,a1
        jsr     (_LVORemove,a6)

        ;------ free our memory
        moveq   #0,d0
        move.l  a5,a1
        move.w  (LIB_NEGSIZE,a5),d0
        sub.l   d0,a1
        add.w   (LIB_POSSIZE,a5),d0
        jsr     (_LVOFreeMem,a6)

        ;------ set up our return value
        move.l  d2,d0
\out
        movem.l   (sp)+,d2/a5/a6
        rts

Null
        moveq   #0,d0
        rts

*** packerinfo - Return pointer to packer info structure
packerinfo:
        lea     (1$,pc),a0
        move.l  a0,d0
        rts

1$      dc.w    1                       ; XpkInfo version
        dc.w    VERSION
        dc.w    0
        dc.w    1                       ; XpkMode version
        dc.l    shortname
        dc.l    longname
        dc.l    descr
        dc.b    'BLZW'
        dc.l    XPKIF_PK_CHUNK!XPKIF_UP_CHUNK!XPKIF_MODES
        dc.l    $7fffffff               ; Max chunk size
        dc.l    $80                     ; Min chunk size
        dc.l    $20000                  ; Default chunk size (128K)
        dc.l    packname
        dc.l    unpackname
        dc.l    packedname
        dc.l    unpackedname
        dc.w    60                      ; Default mode
        dc.w    0
        dc.l    9$                      ; First mode
        ds.l    6

9$      dc.l    10$                     ; Next mode
        dc.l    14
        dc.l    XPKMF_A3000SPEED
        dc.l    3846
        dc.l    2048
        dc.l    159;013
        dc.l    303;424
        dc.w    244
        dc.w    0
        dc.b    "9 bits",0,0

10$     dc.l    11$
        dc.l    28
        dc.l    XPKMF_A3000SPEED
        dc.l    7698
        dc.l    4096
        dc.l    141;598
        dc.l    328;570
        dc.w    294
        dc.w    0
        dc.b    "10 bits",0

11$     dc.l    12$
        dc.l    42
        dc.l    XPKMF_A3000SPEED
        dc.l    15474
        dc.l    8192
        dc.l    135;469
        dc.l    343;764
        dc.w    317
        dc.w    0
        dc.b    "11 bits",0

12$     dc.l    13$
        dc.l    57
        dc.l    XPKMF_A3000SPEED
        dc.l    30882
        dc.l    16384
        dc.l    134;550
        dc.l    356;114
        dc.w    324
        dc.w    0
        dc.b    "12 bits",0

13$     dc.l    14$
        dc.l    71
        dc.l    XPKMF_A3000SPEED
        dc.l    61458
        dc.l    32768
        dc.l    139;276
        dc.l    364;853
        dc.w    329
        dc.w    0
        dc.b    "13 bits",0

14$     dc.l    15$
        dc.l    85
        dc.l    XPKMF_A3000SPEED
        dc.l    122898
        dc.l    65536
        dc.l    143;998
        dc.l    374;032
        dc.w    331
        dc.w    0
        dc.b    "14 bits",0

15$     dc.l    0
        dc.l    100
        dc.l    XPKMF_A3000SPEED
        dc.l    245766
        dc.l    131072
        dc.l    157;748
        dc.l    381;225
        dc.w    337
        dc.w    0
        dc.b    "15 bits",0

*** freebuf - Free the compression table if it's been allocated
* a0 = xbuf
freebuf:
        move.l  (xsp_CompTab,a0),d0
        beq.s   9$
        clr.l   (xsp_CompTab,a0)
        move.l  a6,-(sp)
        move.l  4,a6
        move.l  d0,a1
        move.l  (xsp_CompTabSize,a0),d0
        jsr     (_LVOFreeMem,a6)
        move.l  (sp)+,a6
        moveq   #0,d0
9$      rts

*** getcomptable - Allocate a compression table (if not allocated already or not big enough)
* a3 = xbuf
* d0 = table size needed
* Returns: Z if not enough memory
* a2 = table
getcomptab:
        move.l  (xsp_CompTab,a3),d1
        beq.s   1$
        cmp.l   (xsp_CompTabSize,a3),d0
        bls.s   9$
        move.l  d0,-(sp)
        move.l  a3,a0
        bsr     freebuf
        move.l  (sp)+,d0
1$      move.l  d0,(xsp_CompTabSize,a3)
        moveq   #0,d1
        move.l  a6,-(sp)
        move.l  4,a6
        jsr     (_LVOAllocMem,a6)
        move.l  (sp)+,a6
        move.l  d0,(xsp_CompTab,a3)
        move.l  d0,a2
        rts
9$      move.l  d1,a2
        moveq   #1,d0
        rts

*** packchunk
* a0 = xbuf
* a1 = lochdr
packchunk:
        movem.l d2/a2-a3,-(sp)
        move.l  a0,a3

        move.l  (xsp_Mode,a3),d2        ; Find the compression mode

        mulu.w  #7,d2
        divu.w  #100,d2
        add.w   #9,d2
        ext.l   d2

        moveq   #9,d0
        cmp.l   d0,d2
        bhs.b   .oklow
        move.l  d0,d2
.oklow
        moveq   #15,d0
        cmp.l   d0,d2
        bls.b   .okhigh
        move.l  d0,d2
.okhigh

        move.l  d2,d0                   ; Find a compression table
        bsr     GenCompTableSize
        bsr     getcomptab
        bz.b    \nomem

        move.l  (xsp_InBuf,a3),a0       ; Compress the data
        move.l  (xsp_OutBuf,a3),a1
        move.l  (xsp_InLen,a3),d0
        move.l  d2,d1
        bsr     GenCompress
        move.l  d0,(xsp_OutLen,a3)
        bz.b    \nocomp

        moveq   #0,d0
9$      movem.l (sp)+,d2/a2-a3
        rts

\nomem:
        moveq   #XPKERR_NOMEM,d0
        bra     9$

\nocomp:
        moveq   #XPKERR_EXPANSION,d0
        bra     9$

\nomode:
        moveq   #XPKERR_WRONGMODE,d0
        bra     9$

*** unpackchunk
* a0 = xbuf
unpackchunk:
        movem.l a2-a3,-(sp)
        move.l  a0,a3

        move.l  (xsp_InBuf,a3),a0       ; Find required table size
        bsr     GenDecompTableSize
        bsr     getcomptab
        bz.b    \nomem

        move.l  (xsp_InBuf,a3),a0       ; Decompress the data
        move.l  (xsp_OutBuf,a3),a1
        bsr     GenDecompress
        sub.l   (xsp_OutBuf,a3),a1
        cmp.l   (xsp_OutLen,a3),a1
        bne     \corr

        moveq   #0,d0
9$      movem.l (sp)+,a2-a3
        rts

\nomem:
        moveq   #XPKERR_NOMEM,d0
        bra     9$

\corr:
        moveq   #XPKERR_CORRUPTPKD,d0
        bra     9$

libname         dc.b    "xpkBLZW.library",0

                dc.b    "$VER: "
shortname       dc.b    "BLZW 3.0",0
longname        dc.b    "Bryan's turbo-charged LZW",0
descr           dc.b    "Fast compression and decompression, ratio much like 'compress' or 'zoo'",0
packname        dc.b    "Compressing",0
unpackname      dc.b    "Decompressing",0
packedname      dc.b    "Compressed",0
unpackedname    dc.b    "Decompressed",0

endcode
        end
