/* This is a direct translation of the hardware description.
 * CvR Jul. 1992
 */

#include <stdlib.h>

#include "WordOps.h"
#include "FEAL.h"
#include "FEAL_asm.h"

Word8 __asm
S0 (register __d0 Word8 a, register __d1 Word8 b)
{
  Word16 c;

  c = ((a + b) & 0x00ff) << 2;
  return (Word8) ((c & 0xff) | (c >> 8));
}

Word8 __asm
S1 (register __d0 Word8 a, register __d1 Word8 b)
{
  Word16 c;

  c = ((a + b + 1) & 0x00ff) << 2;
  return (Word8) ((c & 0xff) | (c >> 8));
}

/*
 *  This one needn't be fast, since it's used only for generating the
 *  internal keys from the external one.
 */
Word32
fk (Word32 alpha, Word32 beta)
{
  Word8 a0, a1, a2, a3, b0, b1, b2, b3, f0, f1, f2, f3;

  /* (a0,a1,a2,a3) = alpha */
  a0 = Left8 (Left16 (alpha));
  a1 = Right8(Left16 (alpha));
  a2 = Left8 (Right16(alpha));
  a3 = Right8(Right16(alpha));

  /* (b0,b1,b2,b3) = beta */
  b0 = Left8 (Left16 (beta));
  b1 = Right8(Left16 (beta));
  b2 = Left8 (Right16(beta));
  b3 = Right8(Right16(beta));

  f1 = xor8(a1,a0);
  f2 = xor8(a2,a3);
  f1 = S1(f1, xor8(f2,b0));
  f2 = S0(f2, xor8(f1,b1));
  f0 = S0(a0, xor8(f1,b2));
  f3 = S1(a3, xor8(f2,b3));

  return Concat16(Concat8(f0,f1),Concat8(f2,f3));
}


Word64
password (APTR *s)
{
  Word8* p=(Word8*)s;
  Word64 x;
  Word32 t;

  x.lft = 0;
  x.rgt = 0;

  while (p && *p) {
    t = x.lft;
    x.lft = x.rgt;
    x.rgt = ((t << 9) | (t >> 23)) + (Word32) *p;
    p++;
  }
  return x;
}


void
MakePartKeys (tPartKeys K, Word64 Key, int n)
{
  Word32 A, B, C, D, *k;
  int i;

  k = (Word32*)K;
  A = Left32(Key);
  B = Right32(Key);
  D = 0;

  for (i= n/2 + 4; i>0; i--) {
    C = fk(A, B^D);
    D = A;
    A = B;
    B = C;
    *k++ = C;
  }
}

void
PermPartKeys (tPartKeys K, int n)
{
  int i;
  Word16 T;

  for (i=0; i<n/2; i++) {
    T        = K[n-1-i];
    K[n-1-i] = K[i];
    K[i]     = T;
  }
  for (i=0; i<4; i++) {
    T        = K[n+i];
    K[n+i]   = K[n+4+i];
    K[n+4+i] = T;
  }
}

